/* ttt.c - client-server tic-tac-toe game
 *
 * Copyright © 2005 Carl Worth
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Author: Carl Worth <cworth@cworth.org>
 */

#include <stdio.h>
#include "ttt-board.h"
#include "ttt-error.h"

/* Initialize an empty board. */
void
ttt_board_init (ttt_board_t *board)
{
    int i;
    for (i = 0; i < TTT_BOARD_MAX_CELLS; i++)
    {
	board->cells[i] = TTT_CELL_EMPTY;
    }
    board->current_player = TTT_CELL_X;
}

/* Initialize a board from its string representation.
 *
 * Returns: TTT_BOARD_SUCCESS or TTT_BOARD_INVALID_BOARD if the board
 * string could not be properly parsed.
 */
void
ttt_board_init_from_string (ttt_board_t *board,
			    const char  *s)
{
    int i, j;
    j = 0;
    for (i = 0; i < strlen (s); i++)
    {
	if ((s[i] == 'X') |
	    (s[i] == 'O') |
	    (s[i] == '_'))
	{
	    board->cells[j] = s[i];
	    j++;
	}
    }
}


/* Return the string representation of a board.
 *
 * The return value is a malloc()ed string that should be free()ed
 * when no longer needed.
 *
 * Errors: If out-of-memory occurs, this function will not return.
 */
char *
ttt_board_to_string (ttt_board_t *board)
{
    char *s;
    xasprintf (&s, "\r\n" 
	       "%c|%c|%c\r\n"
	       "%c|%c|%c\r\n"
	       "%c|%c|%c",
	       board->cells[0], board->cells[1], board->cells[2], 
	       board->cells[3], board->cells[4], board->cells[5], 
	       board->cells[6], board->cells[7], board->cells[8]); 
    return s;
}

/* Write a string representation of a board to the provided file.
 *
 * Errors: If out-of-memory or a file IO error occurs, this function
 * will not return.
 */
void
ttt_board_write (ttt_board_t *board, FILE *file)
{
    char *s;

    s = ttt_board_to_string (board);

    xfwrite (s, 1, strlen (s) + 1, file);

    free (s);
}

/* Makes a new move in the given board.
 *
 * The move value is a number from 0 to 9 giving the position to be
 * moved as follows:
 * 
 *  0|1|2
 *  3|4|5
 *  6|7|8
 *
 * Returns: TTT_ERROR_NONE if a move is made
 * succesfully. TTT_ERROR_NOT_VALID_MOVE if a move is invalid (the
 * given cell is not empty). TTT_ERROR_NOT_GRID if the move is
 * impossible (user input 'a' or '9' etc.)
 */

ttt_error_t ttt_board_make_move (ttt_board_t *board, int move)
{  
    if (move < 0 || move > 8)
	return TTT_ERROR_NOT_GRID;
    
    if (board->cells[move] != TTT_CELL_EMPTY)
	return TTT_ERROR_NOT_VALID_MOVE;
    
    board->cells[move] = board->current_player;
    
    if (board->current_player == TTT_CELL_X)
	board->current_player = TTT_CELL_O;	
    else
	board->current_player = TTT_CELL_X;
    return TTT_ERROR_NONE;
}

/* Checks to see if a player has won the game and should be called
 * after every move.
 * 
 * Returns: TTT_CELL_X if X's win. TTT_CELL_O if O's
 * win. TTT_CELL_EMPTY if no win.  The game can then check the value
 * of the return and declare the winner.
 */

ttt_cell_t ttt_board_is_won (ttt_board_t *board)
{
    int i;

    /* Loop over rows and columns looking for a win. */
    for (i = 0; i < 3; i++)
    {
	/* Check the current column */
	if (board->cells[i] != TTT_CELL_EMPTY &&
	    board->cells[i] == board->cells[i + 3] &&
	    board->cells[i] == board->cells[i + 6])
	{
	    return board->cells[i];
	}
	
	/* Check the current row */
	if (board->cells[i * 3] != TTT_CELL_EMPTY &&
	    board->cells[i * 3] == board->cells[(i * 3) + 1] &&
	    board->cells[i * 3] == board->cells[(i * 3) + 2])
	{
	    return board->cells[i * 3];
	}
    }

    if (board->cells[0] != TTT_CELL_EMPTY &&
	board->cells[0] == board->cells[4] &&
	board->cells[0] == board->cells[8])
    {
    	return board->cells[0];
    }

    if (board->cells[2] != TTT_CELL_EMPTY &&
	board->cells[2] == board->cells[4] &&
	board->cells[2] == board->cells[6])
    {
    	return board->cells[2];
    }

    return TTT_CELL_EMPTY;
}
